'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var _construct = _interopDefault(require('@babel/runtime/helpers/construct'));
var _objectWithoutPropertiesLoose = _interopDefault(require('@babel/runtime/helpers/objectWithoutPropertiesLoose'));
var _extends = _interopDefault(require('@babel/runtime/helpers/extends'));
var THREE = require('three');
var React = require('react');
var React__default = _interopDefault(React);
var Reconciler = _interopDefault(require('react-reconciler'));
var omit = _interopDefault(require('lodash-es/omit'));
var upperFirst = _interopDefault(require('lodash-es/upperFirst'));
var ResizeObserver = _interopDefault(require('resize-observer-polyfill'));
var scheduler = require('scheduler');

var roots = new Map();
var emptyObject = {};
function applyProps(instance, newProps, oldProps) {
  if (oldProps === void 0) {
    oldProps = {};
  }

  // Filter equals, events and reserved props
  var sameProps = Object.keys(newProps).filter(function (key) {
    return newProps[key] === oldProps[key];
  });
  var handlers = Object.keys(newProps).filter(function (key) {
    return typeof newProps[key] === 'function' && key.startsWith('on');
  });
  var filteredProps = omit(newProps, [].concat(sameProps, handlers, ['children', 'key', 'ref']));

  if (Object.keys(filteredProps).length > 0) {
    Object.entries(filteredProps).forEach(function (_ref) {
      var key = _ref[0],
          value = _ref[1];
      var root = instance;
      var target = root[key];

      if (key.includes('-')) {
        var entries = key.split('-');
        target = entries.reduce(function (acc, key) {
          return acc[key];
        }, instance);

        if (target && !target.set) {
          // The target is atomic, this forces us to switch the root
          var _entries$reverse = entries.reverse(),
              name = _entries$reverse[0],
              reverseEntries = _entries$reverse.slice(1);

          root = reverseEntries.reverse().reduce(function (acc, key) {
            return acc[key];
          }, instance);
          key = name;
        }
      }

      if (target && target.set) {
        if (target.constructor.name === value.constructor.name) {
          target.copy(value);
        } else if (Array.isArray(value)) {
          var _target;

          (_target = target).set.apply(_target, value);
        } else {
          target.set(value);
        }
      } else root[key] = value;
    });

    if (handlers.length) {
      instance.__handlers = handlers.reduce(function (acc, key) {
        var _extends2;

        var name = key.charAt(2).toLowerCase() + key.substr(3);
        return _extends({}, acc, (_extends2 = {}, _extends2[name] = newProps[key], _extends2));
      }, {}); // Call the update lifecycle, if present

      if (instance.__handlers.update) instance.__handlers.update(instance);
    }
  }
}

function createInstance(type, _ref2) {
  var _ref2$args = _ref2.args,
      args = _ref2$args === void 0 ? [] : _ref2$args,
      props = _objectWithoutPropertiesLoose(_ref2, ["args"]);

  var name = upperFirst(type);
  var instance = type === 'primitive' ? props.object : _construct(THREE[name], args);
  instance.__objects = [];
  instance.__type = type;
  applyProps(instance, props, {});
  return instance;
}

function appendChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.add(child);else {
      parentInstance.__objects.push(child);

      if (child.name) {
        var _applyProps;

        child.__parent = parentInstance;
        applyProps(parentInstance, (_applyProps = {}, _applyProps[child.name] = child, _applyProps));
      }
    }
  }
}

function removeChild(parentInstance, child) {
  if (child) {
    if (child.isObject3D) parentInstance.remove(child);else {
      child.__parent = undefined;
      parentInstance.__objects = parentInstance.__objects.filter(function (c) {
        return c !== child;
      });
    }
  }
}

var Renderer = Reconciler({
  now: scheduler.unstable_now,
  createInstance: createInstance,
  removeChild: removeChild,
  appendChild: appendChild,
  supportsMutation: true,
  isPrimaryRenderer: false,
  schedulePassiveEffects: scheduler.unstable_scheduleCallback,
  cancelPassiveEffects: scheduler.unstable_cancelCallback,
  appendInitialChild: appendChild,
  appendChildToContainer: appendChild,
  removeChildFromContainer: removeChild,
  insertBefore: function insertBefore(parentInstance, child, beforeChild) {
    if (child) {
      if (child.isObject3D) {
        child.parent = parentInstance;
        child.dispatchEvent({
          type: 'added'
        }); // TODO: the order is out of whack if data objects are presents, has to be recalculated

        var index = parentInstance.children.indexOf(beforeChild);
        parentInstance.children = [].concat(parentInstance.children.slice(0, index), [child], parentInstance.children.slice(index));
      } else {
        child.__parent = parentInstance;
        appendChild(parentInstance, child);
      }
    }
  },
  commitUpdate: function commitUpdate(instance, updatePayload, type, oldProps, newProps, fiber) {
    if (instance.isObject3D) {
      applyProps(instance, newProps, oldProps);
    } else {
      // This is a data object, let's extract critical information about it
      var parent = instance.__parent;

      var _newProps$args = newProps.args,
          argsNew = _newProps$args === void 0 ? [] : _newProps$args,
          restNew = _objectWithoutPropertiesLoose(newProps, ["args"]);

      var _oldProps$args = oldProps.args,
          argsOld = _oldProps$args === void 0 ? [] : _oldProps$args,
          restOld = _objectWithoutPropertiesLoose(oldProps, ["args"]); // If it has new props or arguments, then it needs to be re-instanciated


      if (argsNew.some(function (value, index) {
        return value !== argsOld[index];
      })) {
        // First it gets removed from its parent
        removeChild(parent, instance); // Next we create a new instance and append it again

        var newInstance = createInstance(instance.type, newProps);
        appendChild(parent, newInstance); // Switch fiber node

        fiber.alternate.stateNode = newInstance;
        fiber.stateNode = newInstance;
      } else {
        // Otherwise just overwrite props
        applyProps(instance, restNew, restOld);
      }
    }
  },
  getPublicInstance: function getPublicInstance(instance) {
    return instance;
  },
  getRootHostContext: function getRootHostContext(rootContainerInstance) {
    return emptyObject;
  },
  getChildHostContext: function getChildHostContext(parentHostContext, type) {
    return emptyObject;
  },
  createTextInstance: function createTextInstance() {},
  finalizeInitialChildren: function finalizeInitialChildren(instance, type, props, rootContainerInstance) {
    return false;
  },
  prepareUpdate: function prepareUpdate(instance, type, oldProps, newProps, rootContainerInstance, hostContext) {
    return emptyObject;
  },
  shouldDeprioritizeSubtree: function shouldDeprioritizeSubtree(type, props) {
    return false;
  },
  prepareForCommit: function prepareForCommit() {},
  resetAfterCommit: function resetAfterCommit() {},
  shouldSetTextContent: function shouldSetTextContent(props) {
    return false;
  }
});
function render(element, container) {
  var root = roots.get(container);

  if (!root) {
    root = Renderer.createContainer(container);
    roots.set(container, root);
  }

  Renderer.updateContainer(element, root, null, undefined);
  return Renderer.getPublicRootInstance(root);
}
function unmountComponentAtNode(container) {
  var root = roots.get(container);
  if (root) Renderer.updateContainer(null, root, null, function () {
    return roots.delete(container);
  });
}

function useMeasure() {
  var ref = React.useRef();

  var _useState = React.useState({
    left: 0,
    top: 0,
    width: 0,
    height: 0
  }),
      bounds = _useState[0],
      set = _useState[1];

  var _useState2 = React.useState(function () {
    return new ResizeObserver(function (_ref3) {
      var entry = _ref3[0];
      return set(entry.contentRect);
    });
  }),
      ro = _useState2[0];

  React.useEffect(function () {
    if (ref.current) ro.observe(ref.current);
    return function () {
      return ro.disconnect();
    };
  }, []);
  return [{
    ref: ref
  }, bounds];
}

var context = React__default.createContext();
var Canvas = React__default.memo(function (_ref4) {
  var children = _ref4.children,
      style = _ref4.style,
      camera = _ref4.camera,
      renderFn = _ref4.render,
      onCreated = _ref4.onCreated,
      onUpdate = _ref4.onUpdate,
      props = _objectWithoutPropertiesLoose(_ref4, ["children", "style", "camera", "render", "onCreated", "onUpdate"]);

  var canvas = React.useRef();
  var state = React.useRef({
    subscribers: [],
    active: true,
    canvas: undefined,
    gl: undefined,
    camera: undefined,
    scene: undefined,
    size: undefined,
    subscribe: function subscribe(fn) {
      state.current.subscribers.push(fn);
      return function () {
        return state.current.subscribers = state.current.subscribers.filter(function (s) {
          return s === fn;
        });
      };
    }
  });

  var _useMeasure = useMeasure(),
      bind = _useMeasure[0],
      size = _useMeasure[1];

  state.current.size = size;

  var _useState3 = React.useState(function () {
    return new THREE.Raycaster();
  }),
      raycaster = _useState3[0];

  var _useState4 = React.useState(function () {
    return new THREE.Vector2();
  }),
      mouse = _useState4[0];

  var _useState5 = React.useState('default'),
      cursor = _useState5[0],
      setCursor = _useState5[1];

  React.useEffect(function () {
    state.current.scene = window.scene = new THREE.Scene();
    state.current.scene.__objects = [];
    state.current.gl = new THREE.WebGLRenderer({
      canvas: canvas.current,
      antialias: true,
      alpha: true
    });
    state.current.gl.setClearAlpha(0);
    state.current.camera = camera && camera.current || new THREE.PerspectiveCamera(75, 0, 0.1, 1000);
    state.current.gl.setSize(0, 0, false);
    state.current.camera.position.z = 5;
    state.current.canvas = canvas.current;
    if (onCreated) onCreated(state.current);

    var renderLoop = function renderLoop() {
      if (!state.current.active) return;
      requestAnimationFrame(renderLoop);
      if (onUpdate) onUpdate(state.current);
      state.current.subscribers.forEach(function (fn) {
        return fn(state.current);
      });
      if (renderFn) renderFn(state.current);else state.current.gl.render(state.current.scene, state.current.camera);
    }; // Start render-loop


    requestAnimationFrame(renderLoop); // Clean-up

    return function () {
      state.current.active = false;
      unmountComponentAtNode(state.current.scene);
    };
  }, []);
  React.useEffect(function () {
    state.current.gl.setSize(state.current.size.width, state.current.size.height, false);
    var aspect = state.current.size.width / state.current.size.height;
    state.current.camera.aspect = aspect;
    state.current.camera.updateProjectionMatrix();
    state.current.camera.radius = (state.current.size.width + state.current.size.height) / 4;
  });
  var intersect = React.useCallback(function (event, fn) {
    mouse.x = event.clientX / state.current.size.width * 2 - 1;
    mouse.y = -(event.clientY / state.current.size.height) * 2 + 1;
    raycaster.setFromCamera(mouse, state.current.camera);
    var intersects = raycaster.intersectObjects(state.current.scene.children, true);

    for (var i = 0; i < intersects.length; i++) {
      if (!intersects[i].object.__handlers) continue;
      fn(intersects[i]);
    }

    return intersects;
  });
  React.useEffect(function () {
    var hovered = {};

    var handleMove = function handleMove(event) {
      var hover = false;
      var intersects = intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.hover) {
          hover = true;

          if (!hovered[object.uuid]) {
            hovered[object.uuid] = object;
            handlers.hover(data);
          }
        }
      });
      if (hover) cursor !== 'pointer' && setCursor('pointer');else cursor !== 'default' && setCursor('default');
      Object.values(hovered).forEach(function (object) {
        if (!intersects.length || !intersects.find(function (i) {
          return i.object === object;
        })) {
          if (object.__handlers.unhover) object.__handlers.unhover();
          delete hovered[object.uuid];
        }
      });
    };

    window.addEventListener('mousemove', handleMove, {
      passive: true
    });
    return function () {
      window.removeEventListener('mousemove', handleMove);
    };
  }); // Render v-dom into scene

  React.useEffect(function () {
    if (state.current.size.width > 0) {
      render(React__default.createElement(context.Provider, {
        value: _extends({}, state.current),
        children: children
      }), state.current.scene);
    }
  }); // Render the canvas into the dom

  return React__default.createElement("div", _extends({}, bind, props, {
    onClick: function onClick(event) {
      var clicked = {};
      intersect(event, function (data) {
        var object = data.object;
        var handlers = object.__handlers;

        if (handlers.click && !clicked[object.uuid]) {
          clicked[object.uuid] = object;
          handlers.click(data);
        }
      });
    },
    style: {
      cursor: cursor,
      position: 'relative',
      width: '100%',
      height: '100%',
      overflow: 'hidden'
    }
  }), React__default.createElement("canvas", {
    ref: canvas
  }));
});
function useRender(fn) {
  var _useContext = React.useContext(context),
      subscribe = _useContext.subscribe;

  React.useEffect(function () {
    return subscribe(fn);
  }, []);
}
function useThree(fn) {
  var _useContext2 = React.useContext(context),
      subscribe = _useContext2.subscribe,
      props = _objectWithoutPropertiesLoose(_useContext2, ["subscribe"]);

  return props;
}

exports.applyProps = applyProps;
exports.render = render;
exports.unmountComponentAtNode = unmountComponentAtNode;
exports.context = context;
exports.Canvas = Canvas;
exports.useRender = useRender;
exports.useThree = useThree;
